/*
 * Written by Zhen Liu, University of Southern California.
 * $Id: tcp_vegas.c,v 1.3 1995/01/29 00:25:58 zhenliu Exp zhenliu $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/protosw.h>
#include <sys/errno.h>
#include <sys/time.h>

#include <net/route.h>
#include <net/if.h>

#include <netinet/in.h>
#include <netinet/in_pcb.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/tcp_fsm.h>
#include <netinet/tcp_seq.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>
#include <netinet/tcp_vegas.h>
#include <netinet/tcpip.h>
#include <netinet/tcp_debug.h>

/* #define VEGASTRACE 0 */

/*
 * vegastime() returns the time ellapsed (in milliseconds) since
 * 'start', which is supposed to be the starting time of the
 * conversation.
*/
int
vegastime(start)
     	struct timeval start;
{
	struct timeval t;
	uniqtime(&t);
	return ((t.tv_sec - start.tv_sec)*1000 +
		(t.tv_usec - start.tv_usec)/1000);
}

/*
 * Allocate a new vegas stamp from the pool.
 */
struct vstamp *
vstamp_new(vp) 
    	struct vcb *vp; 
{ 
	struct vstamp *new;

	if (vp->q_free == 0)
	    panic("vegas queue full");
	new = vp->q_free;
	vp->q_free = vp->q_free->vs_next;
	return new;
}

/*
 * Return the old stamp to the pool.
 */
void vstamp_free(vp, old)
	struct vcb *vp;
        struct vstamp *old;
{
	/* color the free stamps to help debugging */
	old->vs_len = 9999;
	old->vs_sendtime = 8888;
	old->vs_transmits = 7777;
	/* here is the real work */
	old->vs_next = vp->q_free;
	vp->q_free = old;
}

/*
 * Print out the content of the vegas queue, for debugging.
 */
void 
vegas_print(vp)
	struct vcb *vp;
{
	struct vstamp *vs;
	int i;
	
	printf("\tv_queue=%x, q_free=%x, q_array=%x:\n", 
	       vp->v_queue, vp->q_free, vp->q_array);
	for (i = 1, vs = vp->v_queue; vs != 0; i++, vs = vs->vs_next)
		printf("\t%d: vstamp=%x, len=%d, sendtime=%d, transmits=%d, next=%x\n",
		       i, vs, vs->vs_len, vs->vs_sendtime, vs->vs_transmits, vs->vs_next);
}

      
/*
 * Init various vegas data structure.
 */
void 
vegas_init(vp)
	struct vcb *vp;
{
	int i;

	vp->v_rexmtthresh = 3;
	vp->v_slowstart = 2;		/* New Vegas, mod-2 */

        vp->v_sa = 300 << 3;
        vp->v_sd = 0;
        vp->v_timeout = 1000; /* ((vp->v_sa >> 2) + vp->v_sd) >> 1; */
        vp->v_worried = 0;
        vp->v_time_cwin_chg = 0;
        vp->v_synsent = 0;
        vp->v_send_toggle = 0;
        vp->v_use_sndwnd = 0;  
        vp->v_max_send = 3;
	vp->v_spike_do = 0;		/* default: no spike supression */
	vp->v_spike_nextTime = 0;
	vp->v_spike_timeInc = 1;
	vp->v_spike_byteCnt = 0;
	vp->v_exp_inc_do = 1;		/* default: exp inc */
	vp->v_exp_inc_flag = 1;
	vp->v_exp_inc_cnt = 0;
	vp->v_exp_inc_isless = 0;
	vp->v_exp_inc_nseg = 1;
	vp->v_cong_detect_do = 1;	/* default: congestion detection */
	vp->v_cong_detect_begseq = 0;
	vp->v_cong_detect_begtime = 0;
	
	/* Zhen: This is one of the undocumented feature 
	   of Vegas, it is turned off by default */
	vp->v_cong_detect_predict_do = 0;

	vp->v_cong_detect_baseRtt = 1000000000;
	vp->v_cong_detect_sumRTT = 0;
	vp->v_cong_detect_cntRTT = 0;
	vp->v_cong_detect_top_isless = 0;
	vp->v_cong_detect_top_nseg = 1;	/* default: vegas-1,3 */
	vp->v_cong_detect_bot_isless = 0;
	vp->v_cong_detect_bot_nseg = 3;
	vp->v_incr = 0;

	/* record start time of this session */
	uniqtime(&(vp->v_start));
	
	/* allocate a page of kernel memeory as the pool of
           timestamps - q_array */
	vp->q_array = (struct vstamp *)kmem_alloc(POOLSIZE);
	if (vp->q_array == 0)
	    panic("vegas_init");
	/* thread all timestamps on a free list - q_free */
	vp->q_maxlen = POOLSIZE/sizeof(struct vstamp);
	for (i = 0; i < vp->q_maxlen - 1; i++)
	    vp->q_array[i].vs_next = (struct vstamp *)(vp->q_array + i + 1);
	vp->q_array[vp->q_maxlen-1].vs_next = 0;
	vp->q_free = vp->q_array;

	/* ... and initially the vegas queue is empty */
	vp->v_queue = 0;
}

void 
vegas_spike_on(vp)
	struct vcb *vp;
{
	vp->v_spike_do = 1;
}

void
vegas_exp_inc_off(vp)
	struct vcb *vp;
{
	vp->v_exp_inc_do = 0;
}

void
vegas_cong_detect_off(vp)
	struct vcb *vp;
{
	vp->v_cong_detect_do = 0;
}

void
vegas_24(vp)
	struct vcb *vp;
{
	vp->v_cong_detect_top_nseg = 2;
	vp->v_cong_detect_bot_nseg = 4;
}

void
vegas_cong_detect_pred_on(vp)
	struct vcb *vp;
{
	vp->v_cong_detect_predict_do = 1;
}

/*
 * Called whenever a segment is sent. It create an Vegas timestamp -
 * vstamp, fill in the segment length and time, insert it into the
 * vstamp queue according to the sequence number, truncate the
 * neighboring vstamps if necessary.
 */
 void vegas_send(vp, off, len, currenttime)
	struct vcb *vp;
	int off;
	int len;
	int currenttime;
{
	struct vstamp *vs, *new, *new2, *old;
	int oldoff = off;
	int oldlen = len;

	/* sanity check */
	if (len == 0)
	    goto finish;
	if (off < 0 || len < 0) {
		printf("off=%d, len=%d\n", off, len);
		panic("vegas_send 1");
	}

	/* fill in a new vstamp */
	new = vstamp_new(vp);
	new->vs_len = len;
	new->vs_sendtime = currenttime;
	new->vs_transmits = 1;
	new->vs_next = 0;

	/* if vegas queue is empty, insert the new stamp at the front
           of the queue, then we are done */
	if (vp->v_queue == 0) {
		vp->v_queue = new;
		goto finish;
	}
	/* if offset is zero, insert the new stamp at the front of the
           queue, then keep going */
	if (off == 0) {
		vs = vp->v_queue;
		vp->v_queue = new;
		goto length;
	}
	
	/* ... else from the queue front, count 'off' bytes ... */
	for (vs = vp->v_queue; vs != 0; vs = vs->vs_next) {
	    /* check for 3 cases: 0<off<vs_len, off=vs_len, and off>vs_len */
	    if (off < vs->vs_len) {
		new2 = vstamp_new(vp);
		new2->vs_len = vs->vs_len - off;
		new2->vs_sendtime = vs->vs_sendtime;
		new2->vs_transmits = vs->vs_transmits;
		new2->vs_next = vs->vs_next;
		/* connect the chain, vs->new, and then vs = new2*/
		vs->vs_len = off;
		vs->vs_next = new;
		vs = new2;
		goto length;
	    } else if (off == vs->vs_len) {
		old = vs->vs_next;
		vs->vs_next = new;
		vs = old;
		goto length;
	    } else if (off > vs->vs_len) {
		off -= vs->vs_len;
	    }
	}
	if (vs == 0) {
		printf("oldoff=%d, off=%d\n", oldoff, off);
		panic("vegas_send 2");
	}

length:
	if (vs == 0)
	    goto finish;
	
	/* from now on, the new segment is rexmted segment */
	new->vs_transmits = 2;

	/* now count 'len' bytes ... */
	while (len > 0) {
		/* here is the end of the new segment */
		if (len < vs->vs_len) {
			vs->vs_len = vs->vs_len - len;
			new->vs_next = vs;
			goto finish;
		}
		/* otherwise keep going ... */
		len -= vs->vs_len;
		/* save the pointer to the old stamp, to free it later */
		old = vs;
		vs = vs->vs_next;
		vstamp_free(vp, old);
		/* the end of vegas queue, we are done */
		if (vs == 0)
		    goto finish;
	}
	new->vs_next = vs;

finish:
#ifdef VEGASTRACE
	printf("vegas_send(vp=%x, off=%d, len=%d, currenttime=%ld)\n",
		 vp, oldoff, oldlen, currenttime);
#endif
	return;
}	


/*
 * Called whenever a "good" ACk is received. It deletes the vstamp(s)
 * associated with the sequence space been acked, fragments the last
 * vstamp if necessary. It returned the original send time of the
 * segment(s) (latest), and number of transmissions. Vegas uses this
 * info to compute its fine-grain RTT.
 */
void 
vegas_receive(vp, len, sendtime, transmits)
	struct vcb *vp;
	int len;
	int *sendtime;
	int *transmits;
{
	struct vstamp *vs, *old;
	int oldlen = len;

	*sendtime = 0;
	*transmits = 0;
	vs = vp->v_queue;
	while (len > 0) {
		if (vs == 0) {
			printf("vegas_receive: panic, len=%d\n", len);
			vegas_print(vp);
			panic("vegas_receive top");
		}
		if (*sendtime < vs->vs_sendtime)
		    *sendtime = vs->vs_sendtime;
		if (*transmits < vs->vs_transmits)
		    *transmits = vs->vs_transmits;
		if (len < vs->vs_len) {
			vs->vs_len -= len;
			break;
		}
		len -= vs->vs_len;
		old = vs;
		vs = vs->vs_next;
		vstamp_free(vp, old);
	}
	vp->v_queue = vs;

#ifdef VEGASTRACE
	printf("vegas_receive(vp=%x, len=%d, sendtime=%d, transmits=%d)\n",
		 vp, oldlen, *sendtime, *transmits);
#endif
}

	
/*
 * Go through the vegas queue, see if there is any segment older than
 * 'time'.  Vegas uses it to find out "old" segments which might need
 * to be rexmted, long before the 200ms coarse grain timer timeout.
 */
int
vegas_expire(vp, time, maxlen, sendtime, transmits)
	struct vcb *vp;
    	int time;
        int maxlen;
        int *sendtime;
        int *transmits;
{
	struct vstamp *vs, *old;
	int len = 0;

	if ((vs = vp->v_queue) == 0)
	    return 0;
	*sendtime = 0;
	*transmits = 0;
	if (time < 0) 
	    time = 0;
	while (vs && vs->vs_sendtime < time && len < maxlen) {
		len += vs->vs_len;
		if (*sendtime < vs->vs_sendtime)		
		    *sendtime = vs->vs_sendtime;
		if (*transmits < vs->vs_transmits)
		    *transmits = vs->vs_transmits;
		vs = vs->vs_next;
	}

#ifdef VEGASTRACE
	if (len) {
		printf("%d = vegas_expire(vp=%x, time=%d, maxlen=%d, sendtime=%d, transmits=%d)\n",
		       len, vp, time, maxlen, *sendtime, *transmits);
	}
#endif
	return len;
}
